/**
Action Script library. Color class implementation
Copyright (C) 2005 Christophe Choquet (cchoquet@users.sourceforge.net)  

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_color.hh"
#include "as_sprite.hh"
#include "as_movie_clip.hh"

typedef void (ASColor::*funcPtr)();
static funcPtr callbacks[]= { 
	&ASColor::init,			// 0
	&ASColor::getRGB,		// 1
	&ASColor::getTransform,	// 2
	&ASColor::setRGB,		// 3
	&ASColor::setTransform,	// 4
	// Need to be at the end.
	&ASColor::addProperty,		
	&ASColor::registerClass,	
	&ASColor::AStoString,		
	&ASColor::unwatch,			
	&ASColor::valueOf,			
	&ASColor::watch			
};

/**
 * The constructor
 **/
ASColor::ASColor(Script *s, Graphics *g) : ASObject(s,g,L"Color") {
}

/**
 * The destructor
 **/
ASColor::~ASColor() {
}

/**
 * Return a function descriptor
 */
Function *ASColor::getFunctionMember(wstring memberName) {
	Function *func = getBaseFunction(memberName);
	if (memberName == L"Color")			{func->setFunctionIndex(0); func->setNumParams(1); }
	if (memberName == L"getRGB")		{func->setFunctionIndex(1); func->setNumParams(0); }
	if (memberName == L"getTransform")	{func->setFunctionIndex(2); func->setNumParams(0); }
	if (memberName == L"setRGB")		{func->setFunctionIndex(3); func->setNumParams(1); }
	if (memberName == L"setTransform")	{func->setFunctionIndex(4); func->setNumParams(1); }
	func = finishFunctionSetup(memberName,5,func);
	return func;
}

/**
 * Call class built in functions
 **/
void ASColor::callFunction(Function *func) {
	numPassedParams = func->getNumPassedParams();
	(this->*callbacks[func->getFunctionIndex()])();
}

/**
 * Create a color class. The parameter is a MovieClip name
 **/
void ASColor::init() {
	if (numPassedParams > 0) {
		wstring movieName = script->getRegister(1)->toString();
		if (script->getRegister(1)->getObjectType() == ScriptObjectType && ((ASObject*)script->getRegister(1))->getClassName() == L"Sprite") {
			movieClip = (ASInterface *)script->getRegister(1);
		} else {
			if (movieName.find(L"_root.") != string::npos) movieName = movieName.substr(6);		// Skip the _root. prefix
			movieClip = getRootMovieByName(movieName);
		}
	} else {
		movieClip = NULL;
	}
}

/**
 * A number that represents the RGB numeric value for the color specified.
 **/
void ASColor::getRGB() {
	int rgb = (((color.red <<8)+color.green)<<8)+color.blue;
	script->push(new IntegerObject(rgb));
}

/**
 * Looks it just return a string.
 **/
void ASColor::getTransform() {
	script->push(new StringObject(L"[object Object]"));
}

/**
 * Method; specifies an RGB color for a Color object. Calling this method overrides any previous Color.setTransform()
 **/
void ASColor::setRGB() {
	if (numPassedParams > 0) {
		long l = long(script->getRegister(1)->toNumeric());
		color.red   = (l>>16)&0xff;
		color.green = (l>>8)&0xff;
		color.blue  = l&0xff;
		if (movieClip->getThisObjectInstanceName() == L"Sprite") {
			ASSprite *s = (ASSprite*)movieClip;
			s->changeColor(color);
		}
		if (movieClip->getThisObjectInstanceName() == L"MovieClip") {
			ASMovieClip *s = (ASMovieClip*)movieClip;
			s->changeColor(color);
		}
	}
	noReturnValue();
}

/**
 * Sets color transform information for a Color object. 
 * the Object class must have the following properties that specify color transform values: 
 * ra, rb, ga, gb, ba, bb, aa, ab. 
 **/
void ASColor::setTransform() {
	if (numPassedParams > 0) {
		ASObject *o = (ASObject*) script->getRegister(1);
		int ra,rb, ga, gb, ba, bb, aa, ab;
		if(o->getMember(L"ra") != NULL) transform.redMultTerm   = long(o->getMember(L"ra")->toNumeric()*2.55);
		if(o->getMember(L"rb") != NULL) transform.redAddTerm    = long(o->getMember(L"rb")->toNumeric());
		if(o->getMember(L"ga") != NULL) transform.greenMultTerm = long(o->getMember(L"ga")->toNumeric()*2.55);
		if(o->getMember(L"gb") != NULL) transform.greenAddTerm  = long(o->getMember(L"gb")->toNumeric());
		if(o->getMember(L"ba") != NULL) transform.blueMultTerm  = long(o->getMember(L"ba")->toNumeric()*2.55);
		if(o->getMember(L"bb") != NULL) transform.blueAddTerm   = long(o->getMember(L"bb")->toNumeric());
		if(o->getMember(L"aa") != NULL) transform.alphaMultTerm = long(o->getMember(L"aa")->toNumeric()*2.55);
		if(o->getMember(L"ab") != NULL) transform.alphaAddTerm  = long(o->getMember(L"ab")->toNumeric());
		transform.hasAddTerms = true;
		transform.hasMultTerms = true;
		if (movieClip->getThisObjectInstanceName() == L"Sprite" || movieClip->getThisObjectInstanceName() == L"MovieClip") {
			ASSprite *s = (ASSprite*)movieClip;
			s->setCxFormWithAlpha(&transform);
			s->setHasCxFormWithAlpha(true);
		}
	}
	noReturnValue();
}
