/**
Action Script library. Non AS object. This represents the definition of a Sprite object
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "as_sprite.hh"
#include "../lib/swf_engine.hh"
#include "../lib/event_manager.hh"
#include "as_movie_clip.hh"

/**
 * The constructor
 **/
ASSprite::ASSprite(Script *s, Graphics *g) : ASMovieClip(s,g,L"Sprite")  {
	changeColorRequired = false;
	currentTag = NULL;
	modifiedByScript = false;
	characterDepth = new IntASObjectMap();
	tags = NULL;
	shape = NULL;
}

ASSprite::ASSprite(ASSprite *p) : ASMovieClip(p) {
	currentTag = NULL;
	characterDepth = new IntASObjectMap();
	shape = NULL;
}

/**
 * Deep copy constructor
 **/
ASInterface* ASSprite::getObjectDeepCopy(bool deep) {
	//wcout << L"Enter Copy " << movieName << L"\n";
	ASSprite *newObj = new ASSprite(this);
	newObj->spriteID = spriteID;
	newObj->frameCount = frameCount;
	newObj->tags = tags;
	newObj->depth = depth;
	if (!deep) *(newObj->characterDepth) = *characterDepth;
	newObj->color = color;
	newObj->changeColorRequired = false;
	newObj->mainMember = mainMember;
	if (tags != NULL) newObj->currentTag = tags->begin();
	newObj->modifiedByScript = false;

	if (deep) {
		// Performs a deep copy of inner objects
		for (IntASObjectMap::iterator it=getCharacters()->begin(); it != getCharacters()->end(); ++it) {
			ASObject *amc = (*it).second;
			ASObject *clone = (ASObject*)amc->getObjectDeepCopy(true);
			newObj->_addCharacter((*it).first,clone);
			wstring s = encodeCharacterName(clone->getCharacterId());
			getObjectStore()->setSpriteVariable(newObj,s,clone);
		}
	}
	return newObj;
}

/**
 * The destructor
 **/
ASSprite::~ASSprite() {
	characterDepth->clear();
	for (IntASObjectMap::iterator it=getCharacters()->begin(); it != getCharacters()->end(); it++) delete (*it).second;
}

/**
 * Build a Sprite object from an swf DefineSpriteTag
 **/
void ASSprite::init(DefineSpriteTag *tag) {
	spriteID 	= tag->getSpriteID();
	frameCount 	= tag->getFrameCount();
	tags 		= tag->getTags();
	currentFrame= 0;
	currentTag = tags->begin();
}

/**
 * Go to a dedicated frame inside a script
 **/
void ASSprite::actionGotoFrame(int frameNum) {
	//printf("GOTO FRAME: %d[%x]\n",frameNum,this);
	currentTag = tags->begin();
	for (int i=0; i< frameNum; i++) {
		if ( (*currentTag)->getType() == stagShowFrame) currentTag++;
		while (currentTag != tags->end() && (*currentTag)->getType() != stagShowFrame) {
			int type = (*currentTag)->getType();
			if (type == stagDoInitAction ) {
				script->getEngine()->tagDispatch(*currentTag);
			}
			currentTag++;
		}
	}
	currentFrame = frameNum;
}

/**
 * Return current tag position
 */
list<BaseTag*>::iterator ASSprite::getCurrentTag() {
	return currentTag; 
}

/**
 * Process pending actions before we render the sprite. 
 **/
void ASSprite::toGraphic() {
	if (changeColorRequired) doChangeColor();
	if (needRedraw && shape !=NULL) {
		if (graphics->getSequence(objectUnid) != NULL)	graphics->clearSequence(objectUnid);
		if (graphics->getSequence(objectUnid) == NULL) graphics->processShape(shape,objectUnid,&bounds);
		needRedraw = false;
	}
}

/**
 * Register color changes of all elements in the sprite
 **/
void ASSprite::changeColor(RGB c) {
	changeColorRequired = true;
	color = c;
}

/**
 * Changes the color of all elements in the sprite
 **/
void ASSprite::doChangeColor() {
	for (IntASObjectMap::iterator it = characterDepth->begin(); it != characterDepth->end(); ++it) {
		ASObject* o = (ASObject *)(*it).second;
		if (o->getThisObjectInstanceName() == L"MovieClip") {
			ASMovieClip *m = (ASMovieClip *)o;
			m->changeColor(color);
		} else if (o->getThisObjectInstanceName() == L"Sprite") {
			ASSprite *m = (ASSprite*)o;
			m->changeColor(color);
			m->doChangeColor();
		}

	}
	changeColorRequired = false;
}

void ASSprite::setMainMember(ValueObject  *v){
	setMember(L"MainMember",v);
}

ValueObject* ASSprite::getMainMember() {
	return getMember(L"MainMember");
}

/**
 * Go to next tag
 **/
void ASSprite::goNextTag(bool nextFrame) { 
	//printf("goNextTag: %x  nextVal:%d\n",*currentTag, (*currentTag)->getType()); fflush(stdout);
	currentTag++;
	if (nextFrame) currentFrame++; 
}

int	ASSprite::getCurrentTagType() { 
	if (currentTag == NULL) return 0;
	return (*currentTag)->getType(); 
}

bool ASSprite::isAtEnd() { 
	if (currentTag == NULL) return true;
	return currentTag == tags->end(); 
}


/**
 * Add a character to a sprite AND to all cloned sprites
 **/
void ASSprite::addCharacter(int depth,ASObject *p) {
	// Make sure we are registered as one of the clones
	getObjectStore()->addClonedObject(characterId, this);
	
	// Call update on all clones.
	list<ScriptObject*> l = getObjectStore()->getClonedObjects(characterId);
	for (list<ScriptObject*>::const_iterator it=l.begin(); it != l.end(); ++it) {
		ASSprite *sprite = (ASSprite*) *it;
		if (sprite->getIsStopped() || sprite->getCharacter(depth) != NULL) continue;

		ASObject *tmp = (ASObject *)getObjectStore()->getDeletedObject(sprite,p->getCharacterId());
		if (tmp != NULL) {
			sprite->addCharacterNoRecurs(depth,tmp);			// Restore deleted object
		} else {
			if (sprite == this)	sprite->_addCharacter(depth,p);
			else sprite->_addCharacter(depth,(ASObject *)p->getObjectDeepCopy(true));
		}
	}

	// Register the new object to sprite members
	wstring memberName = p->getMovieName();
	defaultSetMember(memberName,p);
}

void ASSprite::_addCharacter(int depth,ASObject *p) { 
	(*characterDepth)[depth] = p;

	list<ScriptObject*> l = getObjectStore()->getClonedObjects(p->getCharacterId());
	for (list<ScriptObject*>::const_iterator it=l.begin(); it != l.end(); ++it) {
		ASObject *aos = (ASObject*) *it;
		aos->setTransformMatrixClones(p->getTransformMatrix());
		aos->setHasTransformMatrixClones(p->getHasTransformMatrix());
	}
}

void ASSprite::_removeCharacter(int depth) {
	ASObject *oldOne = getCharacter(depth);
	oldOne->setReferenced(false);
	getObjectStore()->addDeletedObject(this,oldOne->getCharacterId(),oldOne);
	characterDepth->erase(depth);
}

void ASSprite::removeCharacter(int depth) {

	// Call update on all clones.
	list<ScriptObject*> l = getObjectStore()->getClonedObjects(characterId);
	for (list<ScriptObject*>::const_iterator it=l.begin(); it != l.end(); ++it) {
		ASSprite *sprite = (ASSprite*) *it;
		if (sprite->getIsStopped()) continue;
		sprite->_removeCharacter(depth);
	}
}

/**
 * Return a Sprite member
 **/
ValueObject*ASSprite::getMember(wstring memberName) {
	//printf("Get property on %x ",this);wcout << memberName << L" " << movieName << L"\n";
	ValueObject *res = getMemberCommon(memberName);
	if (res != NULL) return res;
 	if (res == NULL && memberName == L"_currentframe") {
		res = new IntegerObject(currentFrame +1);
	} else if (memberName == L"_totalframes") {
		res = new IntegerObject(frameCount);		
	} else if (memberName == L"_alpha") { 
		res = new IntegerObject(100);	// It's not CXFORMWITHALPHA. WHat's that alpha value?
	} else if (memberName == L"_width") { 
		v0 =  100000;
		vf = -100000;
		res = new FloatObject(getWidth(this));
	} else if (memberName == L"_height") { 
		v0 =  100000;
		vf = -100000;
		res = new FloatObject(getHeight(this));
	}  else if (memberName == L"_target") { 
		res = new StringObject(L"/"+movieName);	// FIXME: nesting
	} else if (memberName == L"_framesloaded") { 
		res = new IntegerObject(frameCount);	// Frame loaded? Not correct.
	} 	else if (memberName == L"_droptarget") { 
		res = new StringObject(L"");			// ?????
	} else if (memberName == L"_url") { 
		res = new StringObject(L"undefined");	// TODO.
	} else if (memberName == L"_focusrect") { 
		res = new StringObject(L"null");		// Focus rect for buttons
	}  else {
		res = defaultGetMember2(memberName);
	}
	return res;
}

/**
 * Returns the width of a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getWidth(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getWidth((ASSprite*)amc);
		}
		amc->computeBounds();

		float mat1[16],mat2[16], mat3[16];
		convertMatrix(mat2,amc->getTransformMatrix());
		ASObject *tmp = amc;
		while (tmp->getParent() != NULL) {
			tmp = (ASObject*)tmp->getParent();
			if (tmp->getHasTransformMatrix()) {
				convertMatrix(mat1,tmp->getTransformMatrix());
				multiplyMatrix(mat3,mat2,mat1);
				memcpy(mat2,mat3,16*sizeof(float));
			}
		}
		RECT rect = doTransform(*(amc->getBounds()),mat2);
		if (rect.xmin != rect.xmax) {
			float xmin = rect.xmin;
			float xmax = rect.xmax;
			v0 = v0 < xmin ? v0 : xmin;
			vf = vf > xmax ? vf : xmax;
		} 
	}
	return (vf-v0)/20.0;
}

/**
 * Returns the height of a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getHeight(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getHeight((ASSprite*)amc);
		}
		amc->computeBounds();

		float mat1[16],mat2[16], mat3[16];
		convertMatrix(mat2,amc->getTransformMatrix());
		ASObject *tmp = amc;
		while (tmp->getParent() != NULL) {
			tmp = (ASObject*)tmp->getParent();
			if (tmp->getHasTransformMatrix()) {
				convertMatrix(mat1,tmp->getTransformMatrix());
				multiplyMatrix(mat3,mat2,mat1);
				memcpy(mat2,mat3,16*sizeof(float));
			}
		}
		RECT rect = doTransform(*(amc->getBounds()),mat2);
		if (rect.ymin != rect.ymax) {
			float ymin = rect.ymin;
			if(amc->getHasTransformMatrix()) ymin = ymin + amc->getTransformMatrix()->translateY; 
			float ymax = rect.ymax;
			if(amc->getHasTransformMatrix()) ymax = ymax + amc->getTransformMatrix()->translateY; 
			v0 = v0 < ymin ? v0 : ymin;
			vf = vf > ymax ? vf : ymax;
		} 
	}
	return (vf-v0)/20.0;
}

/**
 * Returns the Xmin a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getXmin(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getXmin((ASSprite*)amc);
		}
		RECT *rect = amc->getBounds();
		if (rect->xmin != rect->xmax) {
			v0 = v0 < rect->xmin ? v0 : rect->xmin;
		} 
	}
	return v0/20.0;
}

/**
 * Returns the Xmax a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getXmax(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getXmax((ASSprite*)amc);
		}
		RECT *rect = amc->getBounds();
		if (rect->xmin != rect->xmax) {
			v0 = v0 > rect->xmax ? v0 : rect->xmax;
		} 
	}
	return v0/20.0;
}

/**
 * Returns the Ymin a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getYmin(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getYmin((ASSprite*)amc);
		}
		RECT *rect = amc->getBounds();
		if (rect->ymin != rect->ymax) {
			v0 = v0 < rect->ymin ? v0 : rect->ymin;
		} 
	}
	return v0/20.0;
}

/**
 * Returns the Ymax a sprite
 * Recurse over nested sprites. [A sprite may contains an other, that may contains an other one ......, and so on]
 **/
float ASSprite::getYmax(ASSprite *sprite) {
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); ++it) {
		ASObject *amc = (*it).second;
		if (script->getEngine()->isSpriteObject(amc)) {
			getYmax((ASSprite*)amc);
		}
		RECT *rect = amc->getBounds();
		if (rect->ymin != rect->ymax) {
			v0 = v0 > rect->ymax ? v0 : rect->ymax;
		} 
	}
	return v0/20.0;
}

/**
 * Set a property
 **/
void ASSprite::setMember(wstring memberName, ValueObject *v) {
	//printf("Set property on %x\n",this);

	// Do we have an embedded object that has this property as an object Name?
	for (IntASObjectMap::iterator it=getCharacters()->begin(); it != getCharacters()->end(); ++it) {
		ASSprite *amc = (ASSprite *) (*it).second;
		if (amc->getMovieName() == memberName) {
			amc->setMainMember(v);
		}
	}

	if (memberName == L"_X" || memberName == L"_x") {
		transformMatrix.translateX = int (v->toNumeric() * 20.0);
		hasTransformMatrix = true;
		hasTransformMatrix = true;
	} else if (memberName == L"_Y" | memberName == L"_y") { 
		transformMatrix.translateY = int (v->toNumeric() * 20.0);
		hasTransformMatrix = true;
		hasTransformMatrix = true;
	} else if (memberName == L"_xscale") {
		if (!transformMatrix.hasScale) {
			transformMatrix.scaleX = 65536;
			transformMatrix.scaleY = 65536;
			transformMatrix.hasScale = true;
		}
		double a = v->toNumeric() * 0.01;
		double t = tan(getTransformationMatrixAngle(&transformMatrix)*M_PI/180.0);
		double norm = 1+sqrt(t*t);
		float sign = t > 0 ? 1 : -1;
		double d = 65536 * a  / norm *sign ; 	
		transformMatrix.scaleX = long(d);
		transformMatrix.rotateSkew0 = long(t * d);
		hasTransformMatrix = true;
	} else if (memberName == L"_yscale") { 
		if (!transformMatrix.hasScale) {
			transformMatrix.scaleX = 65536;
			transformMatrix.scaleY = 65536;
			transformMatrix.hasScale = true;
		}
		double a = v->toNumeric() * 0.01;
		double t = tan(getTransformationMatrixAngle(&transformMatrix)*M_PI/180.0);
		double norm = 1+sqrt(t*t);
		float sign = t > 0 ? 1 : -1;
		double d = 65536 * a  / norm *sign ; 	
		transformMatrix.scaleY = long(d);
		transformMatrix.rotateSkew1 = -long(t * d);
		hasTransformMatrix = true;
	} else if (memberName == L"_currentframe") {
	} else if (memberName == L"_totalframes") {
	} else if (memberName == L"_alpha") { 
	} else if (memberName == L"_visible"){
		setVisible(v->toNumeric() != 0.0f);
	} else if (memberName == L"_width") { 
	} else if (memberName == L"_rotation") {
		double angle = v->toNumeric();
		rotateTransformMatrix(&transformMatrix,angle);
		hasTransformMatrix = true;
	} else if (memberName == L"_target") { 
	} else if (memberName == L"_framesloaded") { 
	} else if (memberName == L"_name") {
	} else if (memberName == L"_droptarget") { 
	} else if (memberName == L"_url") { 
	} else if (memberName == L"_highquality") {
	} else if (memberName == L"_focusrect") { 
	} else if (memberName == L"_soundbuftime") { 
	} else if (memberName == L"_quality") { 
	} else if (memberName == L"_xmouse") { 
	} else if (memberName == L"_ymouse") { 
	} else {
		defaultSetMember(memberName,v);
	}
	setDirty();
	modifiedByScript = true;
}

/**
 * Set a variable on a sprite. The set is in the form /_level0/spriteName:member = value
 **/
void ASSprite::setChildrenProperty(wstring member, ValueObject *value) {
	ASSprite *curObj = this;
	bool completed = false;
	while (!completed) {
		// Check if member contains a "/"
		if (member[0] == L'/') member  = member.substr(1);
		wstring subMovieName = member;
		int pos = member.find(L"/");
		if (pos != string::npos) {
			subMovieName = member.substr(0,pos);
			member = member.substr(pos+1);
		} else {
			completed = true;
		}
		for (IntASObjectMap::iterator it=curObj->getCharacters()->begin(); it != curObj->getCharacters()->end(); ++it) {
			ASObject *amc = (*it).second;
			if (amc->getMovieName() == subMovieName) {
				if (completed)	amc->setMainMember(value);
				else {
					curObj = (ASSprite*)amc;
				}
				break;
			}
		}
	}
}


/**
 * returns a movie given its name or its path
 * It's called by object store, when we have nested Sprites.
 **/
ScriptObject*ASSprite::getMovieByNameOrPath(wstring namePath) {
	wcout << namePath << L"\n";
	int po = namePath.find('/');

	ScriptObject*so = NULL;
	if (po == string::npos) {
		for (IntASObjectMap::iterator it=getCharacters()->begin(); it != getCharacters()->end(); ++it) {
			ASObject *amc = (*it).second;
			if (namePath == amc->getMovieName()) so = amc;
		}
	} else {
		wstring main = namePath.substr(0,po);
		for (IntASObjectMap::iterator it=getCharacters()->begin(); it != getCharacters()->end(); ++it) {
			ASSprite *amc = (ASSprite *) (*it).second;
			if (main == amc->getMovieName()) so = amc->getMovieByNameOrPath(namePath.substr(po+1));
		}
	}
	return so;	
}

/**
 * Returns the minimum and maximum x and y coordinates of a SWF file in a specified coordinate space.
 *
 * FIXME implement targetCoordinateSpace
 **/
void ASSprite::ASgetBounds() {
	if (numPassedParams >0 ) { 
		ASObject* targetCoordinateSpace = (ASObject*)script->getRegister(1);
	}
	ASObject* aos = new ASObject(script,graphics);
	v0 =  100000;
	aos->setMemberNoClone(L"xMin",new FloatObject(getXmin(this)));
	v0 =  -100000;
	aos->setMemberNoClone(L"xMax",new FloatObject(getXmax(this)));
	v0 =  100000;
	aos->setMemberNoClone(L"yMin",new FloatObject(getYmin(this)));
	v0 =  -100000;
	aos->setMemberNoClone(L"yMax",new FloatObject(getYmax(this)));
	script->push(aos);
}

/**
 * Duplicates the specified Sprite.
 **/
void ASSprite::duplicateMovieClip() {
	if (numPassedParams > 1 ) {
		wstring name = script->getRegister(1)->toString();
		int depth = int(script->getRegister(2)->toNumeric());
		ASInterface *as = (ASInterface*)script->getCurrentContext();
		ASSprite *newAs = (ASSprite *)as->getObjectDeepCopy(true);
		if (numPassedParams == 3 ) {
			ASObject *initObject = (ASObject *)script->getRegister(3);
			newAs->populateMembersFromInitObject(initObject);
		}
		newAs->setMovieName(name);
		getObjectStore()->setGlobalVariable(name,newAs);
		getObjectStore()->setObjectAtDepth(depth,newAs);
		script->push(newAs);

		ScriptObject *context = script->getCurrentContext();
		script->setCurrentContext(newAs);
		getObjectStore()->targetPathDown(newAs);
		script->getEventManager()->doOnLoad(as);
		getObjectStore()->targetPathUp();
		newAs->setLoaded(true);
		newAs->setIsStopped(false);
		newAs->setIsDuplicated(true);
		script->setCurrentContext(context);
		script->getEventManager()->cloneClipActions(as,newAs);
	} else {
		noReturnValue();
	}
}

/**
 * Sends the playhead to a specific frame in the movie clip and plays the SWF file.
 **/
void ASSprite::gotoAndPlay() {
	if (numPassedParams > 0) {
		int frame = int(script->getRegister(1)->toNumeric());
		actionGotoFrame(frame);
	}
	setIsStopped(false);
	noReturnValue();
}

/**
 * Sends the playhead to a specific frame in the movie clip and stops the SWF file.
 **/
void ASSprite::gotoAndStop() {
	if (numPassedParams > 0) {
		int frame = int(script->getRegister(1)->toNumeric());
		actionGotoFrame(frame);
	}
	setIsStopped(true);
	noReturnValue();
}

/**
 * Returns true if bounding box of the specified movie clip intersects the bounding box of the target movie clip.
 **/
void ASSprite::hitTest() {
	bool res = false;
	if (numPassedParams == 1) {
		ASObject *as = (ASObject*)script->getRegister(1);
	//	RECT r1= doTransform(as->getBounds(),invMatrix);
	//	RECT r2= doTransform(getBounds(),t2); 
	//	res = doesOverlap(, ,as->getTransformMatrix(),getTransformMatrix());
wcout << *as << L" " << res;
	} else if (numPassedParams > 1) {
		float x = script->getRegister(1)->toNumeric();
		float y = script->getRegister(2)->toNumeric();
printf("%f %f\n",x,y);
		puts("TODO");
	}
exit(0);
}

/**
 * Remove a movieclip from a sprite
 **/
void ASSprite::removeMovieClip(){
	if (isDuplicatedObject) {
		setVisible(false);
		setReferenced(false);
		script->getEventManager()->unRegisterClipAction(this);
	}
	noReturnValue();
}

/**
 * Method; creates an empty movie clip as a child of an existing movie clip.
 **/
void ASSprite::createEmptyMovieClip() {
	ASSprite *so = new ASSprite(script,graphics);
	wstring movieName = script->getRegister(1)->toString();
	int  depth = int(script->getRegister(2)->toNumeric());
	so->commonInit();
	so->setMovieName(movieName);
	if (getMovieName() == L"undefined") {
		getObjectStore()->setGlobalVariable(movieName,so);
		getObjectStore()->setObjectAtDepth(depth,so);
	} else {
		_addCharacter(depth,so);
		so->setParent(this);
	}
	so->setMemberNoClone(L"_visible", new BooleanObject(true));
	script->push(so);
}

