/**
Macromedia Flash Maths lib.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "maths.hh"
#include "texture.hh"
#include <list>

/**
 * One D quadratic Bezier interpolation.
 * Two 2D curves just need to call this function of X and Y axis
 **/
float QuadraticBezierInterp(float s, float c, float e, float t) {
	float a = s + (c-s)*t;
	float b = c + (e-c)*t;
	float val = a  + (b-a)*t;
	return val;
} 

/**
 * Return next powerof 2. USe for texture size
 **/
GLuint nextPowerOf2( GLuint in){
	in -= 1;
	in |= in >> 16;
	in |= in >> 8;
	in |= in >> 4;
	in |= in >> 2;
	in |= in >> 1;
	return in + 1;
}

/**
 * Normalise a SWF matrix
 **/
NORMMATRIX normalize(MATRIX *m) {
	int FRAC = (1 << 5);
	NORMMATRIX res;
	res.a 		= double(m->scaleX) / 65536.0;
	res.b 		= double(m->rotateSkew1) / 65536.0;
	res.c 		= double(m->rotateSkew0) / 65536.0;
	res.d 		= double(m->scaleY) / 65536.0;
	res.translateX 	= double(m->translateX);
	res.translateY 	= double(m->translateY);
	return res;
}

/**
 * Inverse a SWF Matrix
 **/
NORMMATRIX MatrixInvert(NORMMATRIX *m) {
	NORMMATRIX res;
	double det;

	det = m->a*m->d-m->b*m->c;
	res.a  = m->d/det;
	res.b  = -m->b/det;
	res.c  = -m->c/det;
	res.d  = m->a/det;

	res.translateX = - (res.a * m->translateX + res.b * m->translateY);
	res.translateY = - (res.c * m->translateX + res.d * m->translateY);
	return res;
}

/**
 * Interpolate an RGBA color (Morphing)
 **/
RGBA interpolateRGBA(RGBA *c1, RGBA *c2, float t) {
	RGBA res;
	res.red = int(c1->red + t*float(c2->red-c1->red));
	res.green = int(c1->green + t*float(c2->green-c1->green));
	res.blue = int(c1->blue + t*float(c2->blue-c1->blue));
	res.alpha = int(c1->alpha + t*float(c2->alpha-c1->alpha));
	return res;
}

/**
 * Interpolate a Matrix (Morphing)
 **/
MATRIX interpolateMATRIX(MATRIX *m1, MATRIX *m2,float t) {
	MATRIX res;
	res.scaleX = long(m1->scaleX + t*float(m2->scaleX-m1->scaleX));
	res.scaleY = long(m1->scaleY + t*float(m2->scaleY-m1->scaleY));
	res.rotateSkew0 = long(m1->rotateSkew0 + t*float(m2->rotateSkew0-m1->rotateSkew0));
	res.rotateSkew1 = long(m1->rotateSkew1 + t*float(m2->rotateSkew1-m1->rotateSkew1));
	res.translateX = long(m1->translateX + t*float(m2->translateX-m1->translateX));
	res.translateY = long(m1->translateY + t*float(m2->translateY-m1->translateY));
	return res;
}

/**
 * Interpolate a Gradient (Morphing)
 **/
GRADIENT *interpolateGRADIENT(GRADIENT *gr1, GRADIENT *gr2,float t) {
	GRADIENT *res = new GRADIENT();
	res->setNumGradients(gr1->getNumGradients());
	list<GRADRECORD*>::iterator it2 = gr2->getGradientRecords()->begin();
	for (list<GRADRECORD*>::iterator it1 = gr1->getGradientRecords()->begin(); it1 != gr1->getGradientRecords()->end(); it1++, it2++) {
		GRADRECORD *g = new GRADRECORD();
		g->setRatio(int((*it1)->getRatio() + t*((*it2)->getRatio()-(*it1)->getRatio())));
		g->setColor(interpolateRGBA((*it1)->getColor(),(*it2)->getColor(),t));
		res->addGradientRecord(g);
	}
	return res;
}

/**
 * Interpolate a Morph Gradient (Morphing)
 **/
GRADIENT *interpolateMORPHGRADIENT(MORPHGRADIENT *m, float t) {
	GRADIENT *res = new GRADIENT();
	res->setNumGradients(m->getNumGradients());
	for (list<MORPHGRADRECORD*>::iterator it = m->getGradientRecords()->begin(); it != m->getGradientRecords()->end(); it++) {
		GRADRECORD *g = new GRADRECORD();
		g->setRatio(int((*it)->startRatio + t*((*it)->startRatio-(*it)->endRatio)));
		g->setColor(interpolateRGBA(&(*it)->startColor,&(*it)->endColor,t));
		res->addGradientRecord(g);
	}
	return res;
}

/**
 * Check if a point is inside a Shape
 **/
bool polygon_contains_point_2d(int nbSegments, Segment *s,float xval,float yval ) {
	float x1;
	float x2;
	float y1;
	float y2;

	bool value = true;
	for (int i = 0; i < nbSegments; i++ ) {
		x1 = s[i].x[0];
		y1 = s[i].y[0];
		x2 = s[i].x[1];
		y2 = s[i].y[1];
		if (s[i].originalIndex[0] != 0 && s[i].originalIndex[1] != 0) {
			if ( ( y1   <  yval && yval <= y2   ) || ( yval <= y1   && y2   < yval ) )  {
				if ( ( xval - x1 ) - ( yval - y1 ) * ( x2 - x1 ) / ( y2 - y1 ) < 0 ) {
					value = !value;
				}
			}
		}
	}
	value = !value;
	return value;
}

/**
 * Check if a point is inside a triangle
 * The routine assumes that the vertices are given in counter-clockwise order.
 **/
bool triangle_contains_point_2d(float v[3][2],float x,float y ) {
  if ( 0.0E+00 < ( x - v[0][0] ) * ( v[1][1] - v[0][1] ) - ( y - v[0][1] ) * ( v[1][0] - v[0][0] ) )  return false;
  if ( 0.0E+00 < ( x - v[1][0] ) * ( v[2][1] - v[1][1] ) - ( y - v[1][1] ) * ( v[2][0] - v[1][0] ) )  return false;
  if ( 0.0E+00 < ( x - v[2][0] ) * ( v[0][1] - v[2][1] ) - ( y - v[2][1] ) * ( v[0][0] - v[2][0] ) )  return false;
  return true;
}

/**
 * Apply transform on a RECT
 **/
RECT doTransform(RECT a, float *m) {
	RECT res;
	res.xmin = long(m[0]*float(a.xmin) + m[4] * float(a.ymin) + m[12]);
	res.ymin = long(m[1]*float(a.xmin) + m[5] * float(a.ymin) + m[13]);
	res.xmax = long(m[0]*float(a.xmax) + m[4] * float(a.ymax) + m[12]);
	res.ymax = long(m[1]*float(a.xmax) + m[5] * float(a.ymax) + m[13]);
	return res;
} 

/**
 * Apply transform on a point
 **/
void doTransform(float &x, float &y, float *m) {
	float xi, yi;
	xi = m[0]*x + m[4] * y + m[12];
	yi = m[1]*x + m[5] * y + m[13];
	x = xi;
	y = yi;
} 

/**
 * Convert a MATRIX struct to a tru transform matrix
 **/
void convertMatrix(float *mat, MATRIX *m) {
	for (int i=0; i< 4; i++) 
		for (int j=0; j <4; j++)
			mat[i*4+j] = i==j  ? 1.0 : 0.0;

	mat[3*4+0] = m->translateX;
	mat[3*4+1] = m->translateY;
	if (m->hasScale ) {
		mat[0*4+0] = (float)m->scaleX/(float)0x10000; 
		mat[1*4+1] = (float)m->scaleY/(float)0x10000;
	}
	if (m->hasRotate) {
		mat[0*4+1] = (float)m->rotateSkew0/(float)0x10000; 
		mat[1*4+0] = (float)m->rotateSkew1/(float)0x10000;
	}
}

/**
 * multiply two transform matrixes
 **/
void multiplyMatrix(float *res, float *mat1, float *mat2) {
	for (int i=0; i< 4; i++) {	
		for (int j=0; j< 4; j++) {
			float r = 0;
			for (int k=0; k< 4; k++) r += mat1[i*4+k]*mat2[k*4+j];
			res[i*4+j] = r;
		}
	}
}

/**
 * Transform a bounding box 
 **/
RECT doTransform(RECT a,MATRIX *m) {
	float mat[16];
	convertMatrix(mat,m);
	return doTransform(a,mat);
}

/**
 * Invert a 4x4 matrix
 * 
 * Code from gamedev.net (yann)
 **/
void invert4x4Matrix(float *data,float *invmat) {
	float r1[8], r2[8], r3[8], r4[8];
	float *s[4] = { r1, r2, r3, r4 };
	float *tmprow;
	float v, scpmax;
	float scp[4];

	int i, j, p, jj, pivot;

	// Setup double temporary matrix from source matrix
	for( i = 0; i < 4; i++ ) for( j = 0; j < 4; j++ ) {
		s[i][j] = data[i*4+j];
		s[i][j+4] = ( i == j ) ? 1.0f : 0.0f;
	}

	for( i = 0; i < 4; i++ ) {
		scp[i] = fabs(s[i][0]);
		for( j = 1; j < 4; j++ ) {
			v = fabs(s[i][j]);
			if( v > scp[i] ) scp[i] = v;
		}
		if( scp[i] == 0.0f ) return ; 			// singular matrix !
	}

	for( i = 0; i < 4; i++ ) {
		// Select pivot row
		pivot = i;
		scpmax = fabs(s[i][i] / scp[i]);

		// Find out which row should be on top
		for( p = i+1; p < 4; p++ ) {
     			v = fabs(s[p][i] / scp[p]);
			if( v > scpmax ) {
				scpmax = v;
				pivot = p; 
			}
		}

		// Pivot if necessary
		if( pivot != i ) {
			tmprow = s[i];
			s[i] = s[pivot];
			s[pivot] = tmprow;
			v = scp[i];
			scp[i] = scp[pivot];
			scp[pivot] = v;
		}		

		// Perform gaussian elimination
		for( j = i+1; j < 4; j++ ) {
			v = s[j][i] / s[i][i];
			s[j][i] = 0.0f;
			for( jj = i+1; jj < 8; jj++ ) s[j][jj] -= v * s[i][jj];
		}
	}

	if( s[3][3] == 0.0f ) return ; 		// singular matrix !

	for( i = 3; i > 0; i-- ) {
		for( j = i-1; j > -1; j-- ) {
			v = s[j][i] / s[i][i];
			for( jj = j+1; jj < 8; jj++ ) s[j][jj] -= v * s[i][jj];
		}
	}

	// Normalize, and copy back to original matrix
	for( i = 0; i < 4; i++ ) for( j = 0; j < 4; j++ ) invmat[i*4+j] = s[i][j+4] / s[i][i];
}

/**
 * Intersect two 2D segments
 **/
bool segmentIntersect(float Ax, float Ay, float Bx, float By, float Cx, float Cy, float Dx, float Dy,float &Px, float &Py) {
	float det1 = (Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx);
	if (det1 == 0) return false;
	float det2 = (Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx);
	if (det2 == 0) return false;
	float r = ((Ay-Cy)*(Dx-Cx)-(Ax-Cx)*(Dy-Cy))/det1;
	float s = ((Ay-Cy)*(Bx-Ax)-(Ax-Cx)*(By-Ay))/det2;
	if (r<0 || r>1 || s<0 || s>1) return false;
	Px=Ax+r*(Bx-Ax);
	Py=Ay+r*(By-Ay);
	return true;	
}

/**
 * Check if two 2D segments are colinear
 **/
bool segmentColinear(float Ax, float Ay, float Bx, float By, float Cx, float Cy, float Dx, float Dy) {
	float v1 = (Bx-Ax)*(Dy-Cy)-(By-Ay)*(Dx-Cx);
	float v2 = (Ay-Cy)*(Dx-Cx)-(Ax-Cx)*(Dy-Cy);
	return (v1 == 0 && v2 == 0);
}


/**
 * Check if the polygon is closed. If not closes the polygon.
 * This is done by checking if all points have two segments comming into
 **/
void closePolygon(Segment *segments,int &nbSegments,float *points, int nbPoints) {
	if (nbPoints <=2 ) return ;
	char *owner = (char*)calloc(nbPoints,1);
	for (int i=0; i< nbSegments; i++) {
		owner[segments[i].points[0]]++;
		owner[segments[i].points[1]]++;
	}

	for (int i=0; i< nbPoints; i++) {
		if (owner[i] < 2) {
			for (int j=0; j< nbPoints; j++) {
				if (i==j || owner[j] >= 2) continue;
				segments[nbSegments].points[0] = i;
				segments[nbSegments].points[1] = j;
				segments[nbSegments].x[0] = points[i*2];
				segments[nbSegments].y[0] = points[i*2+1];
				segments[nbSegments].x[1] = points[j*2];
				segments[nbSegments].y[1] = points[j*2];
				segments[nbSegments].originalIndex[0] = -2;
				segments[nbSegments].originalIndex[1] = -2;
				segments[nbSegments].colorsIndex[0] = 0;
				segments[nbSegments].colorsIndex[1] = 0;
				segments[nbSegments].lineColor = NULL;
				nbSegments++;
			}
		}
	}
}

/**
 * Check if the polygon is closed. 
 **/
bool isClosedPolygon(Segment *segments,int nbSegments, int nbPoints) {
	if (nbPoints <=2 ) return false;
	char *owner = (char*)calloc(nbPoints,1);
	for (int i=0; i< nbSegments; i++) {
		owner[segments[i].points[0]]++;
		owner[segments[i].points[1]]++;
	}
	//for (int i=0; i< nbPoints; i++) printf("%d ",owner[i]); puts("");
	for (int i=0; i< nbPoints; i++) if (owner[i] < 2) return false;

	return true;
}

/**
 * Apply a CxFormWithAlpha upon an other one
 **/
CXFORMWITHALPHA multiplyCxFormWithAlpha(CXFORMWITHALPHA *src1,CXFORMWITHALPHA *src2) {
	CXFORMWITHALPHA res;
	res.redMultTerm = (src1->redMultTerm * src2->redMultTerm) /256;
	res.greenMultTerm = (src1->greenMultTerm * src2->greenMultTerm)/256;
	res.blueMultTerm = (src1->blueMultTerm * src2->blueMultTerm)/256;
	res.alphaMultTerm = (src1->alphaMultTerm * src2->alphaMultTerm)/256;

	res.redAddTerm = (src1->redAddTerm * src2->redMultTerm)/256 + src2->redMultTerm;
	res.greenAddTerm = (src1->greenAddTerm * src2->greenMultTerm)/256 + src2->greenMultTerm;
	res.blueAddTerm = (src1->blueAddTerm * src2->blueMultTerm)/256 + src2->blueMultTerm;
	res.alphaAddTerm = (src1->alphaAddTerm * src2->alphaMultTerm)/256 + src2->alphaMultTerm;
	
	return res;
}

void normalizeTransformMatrix(MATRIX *transformMatrix) {
	// Normalize the transformation matrix
	double norm = (transformMatrix->rotateSkew1*transformMatrix->rotateSkew1+transformMatrix->scaleX*transformMatrix->scaleX);
	norm = sqrt(norm)/65536.0f;
	transformMatrix->rotateSkew0 = long(transformMatrix->rotateSkew0/norm);
	transformMatrix->rotateSkew1 = long(transformMatrix->rotateSkew1/norm);
	transformMatrix->scaleX = long(transformMatrix->scaleX/norm);
	transformMatrix->scaleY = long(transformMatrix->scaleY/norm);
}

/**
 * Angle in degree
 **/
void rotateTransformMatrix(MATRIX *transformMatrix, double angle) {
	angle = fmod(angle,360);
	if (angle < 0) angle = 360+angle;
	double r = angle/180.0*M_PI;
	float a0 = 65536;
	float a1 = 65536;
	if (transformMatrix->hasScale) {
		a0 = transformMatrix->scaleX;
		a1 = transformMatrix->scaleY;
	}

	double initialAngle = getTransformationMatrixAngle(transformMatrix);
	double k = cos(initialAngle*M_PI/180);
	if (fabs(k) > 0.01) {
		a0 = fabs(a0/k);
		a1 = fabs(a1/k);
		transformMatrix->scaleX = long(a0*cos(r));
		transformMatrix->rotateSkew0 = long(sin(r)*a0);
		transformMatrix->rotateSkew1 = -long(sin(r)*a1);
		transformMatrix->scaleY = long(a1*cos(r));
	} else {
		a0 = 52000;
		a1 = 52000;
		transformMatrix->scaleX = long(a0*cos(r));
		transformMatrix->rotateSkew0 = long(sin(r)*a0);
		transformMatrix->rotateSkew1 = -long(sin(r)*a1);
		transformMatrix->scaleY = long(a1*cos(r));
	}
	transformMatrix->hasRotate = true;
	transformMatrix->hasScale = true;
}

/**
 * Return transformation maxtrix angle in degree
 **/
double getTransformationMatrixAngle(MATRIX *transformMatrix) {
	double r=0;
	///printf("-> %d %d\n",transformMatrix->rotateSkew0,transformMatrix->scaleX);
	if (transformMatrix->hasRotate) {
		float a0 = 65536;	
		float a1 = 65536;	
		if (transformMatrix->hasScale) {
			a0 = transformMatrix->scaleX;
			a1 = transformMatrix->scaleY;
		}
		double a = atan(transformMatrix->rotateSkew0/a0);
		if (a0 >= 0) {
			r = a * 180 / M_PI;
			if (r < 0) r = 360 + r;
		}
		else r = 180 + a * 180 / M_PI;
	} else {
		if (transformMatrix->scaleX < 0) r = 180;
	}
	return r;
}

/**
 * Test if two RECT overlaps
 **/
bool doesOverlap(RECT *r1, RECT *r2) {
	bool res = false;
	if (r1->xmin < r2->xmin && r1->xmax < r2->xmin) return false;
	if (r1->xmax > r2->xmin && r1->xmin < r2->xmin) return false;
puts("TOFINISH!!!!!!!!!!!!!!!!!!");
}
