/**
Macromedia Flash OpenGl renderer engine.
Copyright (C) 2005  Christophe Choquet (cchoquet@users.sourceforge.net)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**/
#include "swf_engine.hh"
#include "maths.hh"
#include "uncompress.hh"
#include "ASLib/as_movie_clip.hh"
#include "ASLib/as_text_field.hh"
#include "ASLib/as_video.hh"
#include "ASLib/as_sound.hh"
#include "ASLib/as_selection.hh"
#include "event_manager.hh"
#include "swf_decoder.hh"
#include "localFonts.hh"

extern bool debug;

/**
 * The constructor
 **/
SwfEngine::SwfEngine(Swf_decoder *decoder) {
	this->tagList = decoder->getTagList();
	header = *(decoder->getHeader());
	mainTag = tagList->begin();
	texture = new Texture();
	oglRenderer = new OglRenderer(texture);
	localFontManager = new LocalFontManager(texture);
	graphics = new Graphics(texture,localFontManager);
	objectStore = new ObjectStore(graphics);
	script = new Script(objectStore,graphics, this);
	eventManager = new EventManager(objectStore,graphics, this);
	drawingInProcess = false;
	pthread_mutex_init(&synchro, NULL);
	highestMaskedLayer = 0;
	executeMode = true;
	currentFrame = 0;
	global = new ASObject(script,graphics,L"_global");
	global->setIsStopped(false);
	global->setMovieName(L"_global");
	objectStore->setGlobalVariable(L"_global",global);	
	noActionScript = false;
	actionSkip = 0;
	waitFrame = 0;
	actionListChanged = false;
}

/**
 * The destructor
 **/
SwfEngine::~SwfEngine() {
	pthread_mutex_lock(&synchro);
	drawingInProcess = true;
	pthread_mutex_unlock(&synchro);
	delete oglRenderer;
	delete localFontManager;
	delete texture;
	delete objectStore;
	delete script;
	delete eventManager;
	delete graphics;
}

/**
 * Build next frame, if any available
 * Call the decoding until we reach a stagShowFrame tag
 * It's a three steps process:
 * - 1- process InitActions
 * - 2- process non ActionScript tags
 * - 3- process AS tags
 * That's because an action script may change the properties of a non AS object. Like I define a TextArea graphically and then I change the text
 * by an action script command.
 **/
void SwfEngine::next() {
	if (waitFrame) {
		if (waitFrame >= currentFrame) {
			waitFrame = 0;
		} else {
			currentFrame++;
			waitFrame --;
			return;
		}
	}
	if (mainTag == tagList->end() && header.frameCount > 1) {
		mainTag = tagList->begin();
		script->getObjectStore()->clear();
		doActionList.clear();
	}
	if (getGlobal()->getIsStopped() && executeMode) return;

	// Process initActions;
	list<BaseTag*>::iterator tmpTag = mainTag;
	if ( tmpTag != tagList->end()) {
		if ( (*tmpTag)->getType() == stagShowFrame) tmpTag++;
		while (tmpTag != tagList->end() && (*tmpTag)->getType() != stagShowFrame) {
			int type = (*tmpTag)->getType();
			if (type == stagDoInitAction ) {
				tagDispatch(*tmpTag);
			}
			tmpTag++;
		}  
	}

	// Process graphic Only tags
	if ( mainTag != tagList->end()) {
		if ( (*mainTag)->getType() == stagShowFrame) { currentFrame++; mainTag++; }
		while (mainTag != tagList->end() && (*mainTag)->getType() != stagShowFrame) {
			int type = (*mainTag)->getType();
			if (type != stagDoInitAction) {
				tagDispatch(*mainTag);
			}
			mainTag++;
		}
	}
	doShowFrame();
}


/**
 * Calls the renderer (OpenGl) to display things
 * Characters are rendered in the reverse order.
 * This func is protected from beeing called when the previous frame is still beeing rendered (slow hardware).
 * During rendering, deleting the SwfEngine object is protected by mutex (because rendering & frame management is asynchronous).
 **/
void SwfEngine::display() {
	if (drawingInProcess) return;		
	drawingInProcess = true;
	pthread_mutex_lock(&synchro);
	oglRenderer->clearWindow();
	highestMaskedLayer = 0;

	doDisplay();
	drawingInProcess = false;
	pthread_mutex_unlock(&synchro);
}

/**
 * Performs the character rendering
 **/
void SwfEngine::doDisplay() {
	//puts("\nbegin dodisplay");

	int currentDepth = 0;
	IntScriptObjectMap *movies = script->getObjectStore()->getObjectByDepth();
	for (IntScriptObjectMap::iterator it=movies->begin(); it != movies->end(); ++it) {
		ASSprite *obj = (ASSprite*)(*it).second;
		if (obj == NULL) continue;
		currentDepth = (*it).first;
		oglRenderer->initCharacter();
		if (obj->getHasTransformMatrix()) oglRenderer->applyMatrix(obj->getTransformMatrix());
		if (obj->getHasCxFormWithAlpha()) oglRenderer->pushCxFormWithAlpha(obj->getCxFormWithAlpha());
		if (isSpriteObject(obj)) {
			innerDoDisplay((ASSprite*)obj);
		} else {
			if (currentDepth > highestMaskedLayer) oglRenderer->resetClipDepth();
			if (obj->getClipDepth() != 0) highestMaskedLayer = obj->getClipDepth();
			if (obj->isVisible()) {
				if (obj->getThisObjectInstanceName() == L"Button") {
					obj->toGraphic();
					renderButton((ASButton *)obj,currentDepth);
				} else {
					//wcout << L"Render Depth:" << currentDepth << L" objectName:" <<movieName << L" highestMaskedLayer:" << highestMaskedLayer << L"\n";
					obj->toGraphic();
					eventManager->updateSelectionArea(currentDepth,obj, L"");
					oglRenderer->renderCharacter(graphics->getSequence(obj->getUnid()),obj->getClipDepth());
				}
			}
		}
		if (obj->getHasCxFormWithAlpha()) oglRenderer->popCxFormWithAlpha();
	}
	ASSelection *as = (ASSelection*)script->getObjectStore()->getGlobalVariable(L"_SelectionSingleton");
	if (as != NULL) {	
		if (as->isCaretActive()) oglRenderer->displayCaret(as->getCaret());
	}
	oglRenderer->flush();
	//puts("\nend dodisplay");
}

/**
 * Render a button. A button has a list of active Shapes/Characters/Text edit areas
 **/
void SwfEngine::renderButton(ASButton *amc, int currentDepth) {
	//wcout << L"Button\n";
	for (list<BUTTONRECORD*>::iterator it = amc->getActiveCharacters()->begin();it != amc->getActiveCharacters()->end(); it++) {
		oglRenderer->pushMatrix();
		BUTTONRECORD* b= *it;
		wstring movieName = encodeCharacterName(b->characterID);
		if (b->placeDepth) oglRenderer->applyMatrix(&(b->placeMatrix));
		if (b->hasColorTransform) oglRenderer->pushCxFormWithAlpha(&(b->colorTransform));
		ASInterface *a = (ASInterface *) objectStore->getSpriteVariable(amc,movieName);
		if (b->buttonStateHitTest) {
			eventManager->updateSelectionArea(currentDepth,amc,movieName);
		}
		if (b->buttonStateDown || b->buttonStateOver || b->buttonStateUp) {
			a->toGraphic();
			oglRenderer->renderCharacter(graphics->getSequence(a->getUnid()),a->getClipDepth());
		}
		if (b->hasColorTransform) oglRenderer->popCxFormWithAlpha();
		oglRenderer->popMatrix();
	}
}

/**
  * Recursive calls to character rendering in sprites
  **/
void SwfEngine::innerDoDisplay(ASSprite *sprite) {
	if (!sprite->isVisible()) return;
	//printf("Render %x %d %d\n",sprite,sprite->getCurrentFrame(),sprite->getCharacterId());
	objectStore->targetPathDown(sprite);
	sprite->toGraphic();
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); it++) {
		int depth = (*it).first;
		ASObject *obj = (*it).second;
		oglRenderer->pushMatrix();
		if (obj->getHasTransformMatrix()) oglRenderer->applyMatrix(obj->getTransformMatrix());  
			else oglRenderer->noMatrix();
		if (obj->getHasCxFormWithAlpha()) oglRenderer->pushCxFormWithAlpha(obj->getCxFormWithAlpha());
		if (isSpriteObject(obj)) {
			innerDoDisplay((ASSprite*)obj);
		} else {
			//wcout << L"Render Sprite Depth:"<< depth << L" S/shapeName:" << sprite->getMovieName() << L"\n"; //%s  %x clipMask:%d\n"highestMaskedLayer);
			if (depth > highestMaskedLayer) oglRenderer->resetClipDepth();
			if (obj->getClipDepth() != 0) highestMaskedLayer = obj->getClipDepth();
			if (obj->getThisObjectInstanceName() == L"Button") {
				obj->toGraphic();
				renderButton((ASButton *)obj,depth);
			} else {
				obj->toGraphic();
				eventManager->updateSelectionArea(depth,obj,L"");
				oglRenderer->renderCharacter(graphics->getSequence(obj->getUnid()),obj->getClipDepth());
			}
		}
		if (obj->getHasCxFormWithAlpha()) oglRenderer->popCxFormWithAlpha(); 
		oglRenderer->popMatrix();
	}
	objectStore->targetPathUp();
}

/**
 * Dispatch tag execution
 **/
void SwfEngine::tagDispatch(BaseTag *tag) {
	int code = tag->getType();
	//wcout << "Tag: "<< code << L"\n";
	GLint newDisplayListEntry = -1;
	bool decoded = false;
	switch (code) {
		case stagEnd: {
			decoded = true;
			break;
		}
		case stagShowFrame: {
			decoded = true;
			doShowFrame();
			break;
		}
		case stagDefineShape: {
			decoded = true;
			DefineShapeTag* t = (DefineShapeTag*)tag ;
			wstring movieName = encodeCharacterName(t->getShapeId());
			ASMovieClip *amc = (ASMovieClip *) objectStore->getGlobalVariable(movieName);		
			if (amc == NULL) {
				amc = new ASMovieClip(script,graphics);
				amc->setShape(t->getShapeWithStyle());
				amc->setBounds(t->getShapeBounds());
				amc->setCharacterId(t->getShapeId());
				amc->setMovieName(movieName);
				objectStore->setGlobalVariable(movieName,amc);
				if (debug) wcout << L"+Shape " << movieName << L"TagSpriteId:" << t->getCurSpriteId() << L"\n";
			} else {
				amc->setReferenced(false);
			}
			break;
		}
		case stagFreeCharacter:	 		break;
		case stagPlaceObject: 			break;
		case stagRemoveObject: 			break;
		case stagDefineBits: {
			decoded = true;
			DefineBitsTag *t = (DefineBitsTag*)tag;
			if (!texture->bitmapDefined(t->getCharacterID())) {
				Image *img = new Image();
				UncompressTools *tools = new UncompressTools();
				bool res = tools->buildFromJpegAbbreviatedData(t->getJPEGData(),jpegTableData,img);
				if (res) texture->registerExternalBitmap(t->getCharacterID(),img);
				delete img;
				delete tools;
				if(debug) wcout << L"+Bits "<< t->getCharacterID() << L"\n";
			}
 			break;
		}
		case stagDefineButton: 			break;
		case stagJPEGTables: {
			decoded = true;
			JPEGTablesTag *t = (JPEGTablesTag*)tag;
			jpegTableData = t->getJPEGData();
			break;
		}
		case stagSetBackgroundColor:{
			decoded = true;
			oglRenderer->setBackGroundColor(((SetBackgroundColorTag*)tag)->getColor());
			graphics->setBgColor(((SetBackgroundColorTag*)tag)->getColor());
			break;
		}
		case stagDefineFont:{
			decoded = true;
			DefineFontTag* t = (DefineFontTag*) tag;
			if (texture->characterDefined(t->getFontID())) {
				localFontManager->removeFont(t->getFontID());
			}
			UI16 *offsetTable = t->getOffsetTable();
			int fontId = t->getFontID();
			int nbShape = offsetTable[0];
			list<UI8*> bitmaps;
			for (int i=0; i< nbShape; i++) {
				Shape *shape = &t->getGlyphShapeTable()[i];
				graphics->currentShapeIsCharacter();
				graphics->processShape(shape,t->getFontID(),NULL);
				bitmaps.push_back(graphics->getRaster());
			}
			localFontManager->loadGlyphFaces(fontId,&bitmaps);
			list<UI8*>::iterator it;
			for (it = bitmaps.begin(); it != bitmaps.end(); it++) {
				delete []*it;
				*it = NULL;
			}
			graphics->clearRaster();
			bitmaps.clear();
			graphics->currentShapeIsNotCharacter();
			if(debug) wcout << L"+Font "<< t->getFontID() << L"\n";
 			break;
		}
		case stagDefineText:{
			decoded = true;
			DefineTextTag *t = (DefineTextTag*)tag;
			ASTextField *atf = (ASTextField *)objectStore->getGlobalVariable(encodeCharacterName(t->getCharacterID()));
			if (atf == NULL) {
				atf = new ASTextField(script,graphics);
				atf->setStaticTextField(t);
				atf->setBounds(t->getTextBounds());
				atf->toGraphic();
				atf->setCharacterId(t->getCharacterID());
				objectStore->setGlobalVariable(encodeCharacterName(t->getCharacterID()),atf);
				if(debug) wcout << L"+Text "<< t->getCharacterID() << L"\n";
			} else {
				atf->setReferenced(false);
			}
			break;
		}
		case stagDoAction: {
			decoded = true;
			DoActiontag *t = (DoActiontag*) tag;
			if (!noActionScript) {
				bool alreadyDefined = false;
				for (list<DoActiontag*>::const_iterator it = doActionList.begin(); it != doActionList.end(); ++it) alreadyDefined = *it == tag;
				if (!alreadyDefined) doActionList.push_back(t);	// We don't want action script to execute inside a GotoFrame
			}
			break;
		}
		case stagDefineFontInfo: {
			decoded = true;
			DefineFontInfoTag *t = (DefineFontInfoTag*) tag;
			if (!texture->characterDefined(t->getFontID())) {
				if (localFontManager->getChararacterTexture(t->getFontID(),0) == NULL) {
					localFontManager->loadFontFaces((char*)t->getFontName(),t->getFontID(),t->getFontFlagsBold(),
						t->getFontFlagsItalic(),t->getCodeTableLen(),t->getCodeTable());
				}
			}
		 	break;
		}
		case stagDefineSound: {
			decoded = true;
			ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
			if (as != NULL) { // Should never be null
				DefineSoundTag *t = (DefineSoundTag*)tag;
				as->createSoundTrack(t);
			}
		}
		break;
		case stagStartSound:  {
			decoded = true;
			ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
			if (as != NULL) { // Should never be null
				StartSoundTag*t = (StartSoundTag*)tag;
				as->startSound(t->getSoundId(),t->getSoundInfo());
			}
		}
		case stagStopSound: 		 		break;
		case stagDefineButtonSound:  break;
		case stagSoundStreamHead: 	  {
			decoded = true;
			SoundStreamHeadTag *t = (SoundStreamHeadTag*)tag;
			ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
			if (as == NULL) {
				as = new ASSound(script,graphics);
				as->init(t);
				if (debug) wcout << L"+Create Sound manager\n";
				objectStore->setGlobalVariable(L"_SoundManager",as);
			}
			break;
		}
		case stagSoundStreamBlock: {
			decoded = true;
			ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
			if (as != NULL) as->processSoundStreamBloc((SoundStreamBlockTag*)tag);
			break;
		}
		case stagDefineBitsLossless: {
			decoded = true;
			DefineBitsLosslessTag *t = (DefineBitsLosslessTag*)tag;
			if (!texture->bitmapDefined(t->getCharacterID())) {
				Image *img = new Image();
				UncompressTools *tools = new UncompressTools();
				bool res = tools->buildFromZlibData(t->getZlibBitmapData(),img,t->getBitmapWidth(),t->getBitmapHeight(),t->getBitmapFormat(),
												t->getBitmapColorTableSize(),false);
				if (res) texture->registerExternalBitmap(t->getCharacterID(),img);
				delete img;
				if(debug) wcout << L"+BitsLossless "<< t->getCharacterID() << L"\n";
			}
			break;
		}
		case stagDefineBitsJPEG2:{
			decoded = true;
			DefineBitsJPEG2Tag *t = (DefineBitsJPEG2Tag*)tag;
			if (!texture->bitmapDefined(t->getCharacterID())) {
				Image *img = new Image();
				UncompressTools *tools = new UncompressTools();
				bool res = tools->uncompressJPEGData(t->getJPEGData(),img,false,0);
				if (res) texture->registerExternalBitmap(t->getCharacterID(),img);
				delete img;
				delete tools;
				if(debug) wcout << L"+DefineBitsJPEG2 "<< t->getCharacterID() << L"\n";
			}
 			break;
		}
		case stagDefineShape2: 	break;          // All Shape Define2 are processed as Shape Define
		case stagDefineButtonCxform:break;
		case stagProtect: {
			decoded = true;
			break;
		}
		case stagPlaceObject2:{
			decoded = true;
			PlaceObject2Tag*t = (PlaceObject2Tag*)tag;
			doPlaceObject2(t);
 		 	break;
		}
		case stagRemoveObject2: {
			decoded = true;
			removeObject((RemoveObject2Tag*) tag);
	 	 	break;
		}

		case stagDefineShape3: break;		// All Shape Define3 are processed as Shape Define
		case stagDefineText2: {
			decoded = true;
			DefineText2Tag*t = (DefineText2Tag*)tag;
			ASTextField *atf = (ASTextField *)objectStore->getGlobalVariable(encodeCharacterName(t->getCharacterID()));
			if (atf == NULL) {
				atf = new ASTextField(script,graphics);
				atf->setStaticTextField(t);
				atf->setBounds(t->getTextBounds());
				atf->toGraphic();
				atf->setCharacterId(t->getCharacterID());
				objectStore->setGlobalVariable(encodeCharacterName(t->getCharacterID()),atf);
				if(debug) wcout << L"+DefineText2 "<< t->getCharacterID() << L"\n";
			} else {
				atf->setReferenced(false);
			}
	 		break;
		}
		case stagDefineButton2: {
			decoded = true;
			DefineButton2Tag *t = (DefineButton2Tag*)tag;
			ASButton *but = (ASButton*)objectStore->getGlobalVariable(encodeCharacterName(t->getButtonId()));
			if (but == NULL) {
				but = new ASButton(script,graphics);
				but->setButtton(t);
				but->toGraphic();
				but->setCharacterId(t->getButtonId());
				wstring bname = encodeCharacterName(t->getButtonId());
				objectStore->setGlobalVariable(bname,but);
				if (debug) wcout << L"+Button "<< bname << L" TagSpriteId:" << t->getCurSpriteId() << L"\n";
			} else {
				but->setReferenced(false);
			}
		 	break;
		}

		case stagDefineBitsJPEG3:  {
			decoded = true;
			DefineBitsJPEG3Tag *t = (DefineBitsJPEG3Tag*)tag;
			if (!texture->bitmapDefined(t->getCharacterID())) {
				Image *img = new Image();
				UncompressTools *tools = new UncompressTools();
				bool res = tools->uncompressJPEGData(t->getJPEGData(),img,true,t->getAlphaDataOffset());
				if (res) texture->registerExternalBitmap(t->getCharacterID(),img);
				delete img;
				delete tools;
				if(debug) wcout << L"+BitsJPEG3 "<< t->getCharacterID() << L"\n";
			}
			break;
		}
		case stagDefineBitsLossless2: {
			decoded = true;
			DefineBitsLossless2Tag *t = (DefineBitsLossless2Tag*)tag;
			if (!texture->bitmapDefined(t->getCharacterID())) {
				Image *img = new Image();
				UncompressTools *tools = new UncompressTools();
				bool res = tools->buildFromZlibData(t->getZlibBitmapData(),img,t->getBitmapWidth(),t->getBitmapHeight(),
							t->getBitmapFormat(), t->getBitmapColorTableSize(), true);
				if (res) texture->registerExternalBitmap(t->getCharacterID(),img);
				delete tools;
				if(debug) wcout << L"+BitsLossless2 "<< t->getCharacterID() << L"\n";
			}
			break;
		}
		case stagDefineEditText: {
			decoded = true;
			DefineEditTextTag *t = (DefineEditTextTag*)tag;
			ASTextField *atf = new ASTextField(script,graphics);
			atf->setDynamicText(t);
			atf->toGraphic();
			atf->setCharacterId(t->getCharacterID());
			atf->setMovieName(t->getVariableName()->getString());
			wstring s = t->getVariableName()->getString();
			objectStore->setGlobalVariable(s,atf);
			objectStore->setGlobalVariable(encodeCharacterName(t->getCharacterID()),atf);
			if(debug) wcout << L"+EditText "<< t->getCharacterID() << L"\n";
			break;
		}
		case stagDefineSprite: {
			decoded = true;
			DefineSpriteTag *t = (DefineSpriteTag*)tag;
			wstring movieName = encodeCharacterName(t->getSpriteID());
			ASSprite *as = (ASSprite*)objectStore->getGlobalVariable(movieName);
			if (as == NULL) {
				as = new ASSprite(script,graphics);
				as->init(t);
				as->setCharacterId(t->getSpriteID());
				as->setMovieName(movieName);
				if (debug) wcout << L"+Sprite " << movieName << L"  TagSpriteId:" << tag->getCurSpriteId() << L"\n";
				objectStore->setGlobalVariable(movieName,as);
				bool saveNoActionScript = noActionScript;
				noActionScript = false;
				innerDisplay(as);
				as->actionGotoFrame(0);
				noActionScript = saveNoActionScript;
			} else {
				as->setReferenced(false);
			}
			break;
		}
		case stagNameCharacter: 	 	break;
		case stagFrameLabel: {
			decoded = true;
			FrameLabelTag *t = (FrameLabelTag*)tag;
			script->setLabel(t->getName()->getString(),currentFrame);
			break;
		}
		case stagSoundStreamHead2: {
			decoded = true;
			SoundStreamHead2Tag *t = (SoundStreamHead2Tag*)tag;
			ASSound *as = (ASSound*)objectStore->getGlobalVariable(L"_SoundManager");
			if (as == NULL) {
				as = new ASSound(script,graphics);
				as->init(t);
				if (debug) wcout << L"+Create Sound manager\n";
				objectStore->setGlobalVariable(L"_SoundManager",as);
			}
			break;
		}
		case stagDefineMorphShape: 	{
			decoded = true;
			DefineMorphShapeTag* t = (DefineMorphShapeTag*)tag;
			ASMovieClip *amc = (ASMovieClip*)objectStore->getGlobalVariable(encodeCharacterName(t->getCharacterID()));
			if (amc == NULL) {
				amc = new ASMovieClip(script,graphics);
				amc->setMorphShape(t);
				amc->setCharacterId(t->getCharacterID());
				wstring movieName = encodeCharacterName(t->getCharacterID());
				objectStore->setGlobalVariable(movieName,amc);
				if (debug) wcout << L"+MorphShape " << movieName << L"\n";
			} else {
				amc->setReferenced(false);
			}
			break;
		}
		case stagDefineFont2: {
			decoded = true;
			DefineFont2Tag* t = (DefineFont2Tag*)tag;
			if (!texture->characterDefined(t->getFontID())) {
				if (texture->characterDefined(t->getFontID())) {
					localFontManager->removeFont(t->getFontID());
				}
				bool hasLocalFont;
				if (t->getNumGlyphs() > 0) {
					//hasLocalFont = localFontManager->loadFontFaces((char*)t->getFontName(),t->getFontID(),
					//	t->getFontFlagsBold(),t->getFontFlagsItalic(),t->getNumGlyphs(),t->getCodeTable());
					// Prefer using embedded fonts - They are more accurate.
					hasLocalFont = false;
				} else {
					hasLocalFont = localFontManager->loadFullFontFaces((char*)t->getFontName(),t->getFontID(),
									t->getFontFlagsBold(),t->getFontFlagsItalic());
				}
				if (!hasLocalFont) {
					int fontId = t->getFontID();
					int nbShape = t->getNumGlyphs();
					list<UI8*> bitmaps;
					for (int i=0; i< nbShape; i++) {
						Shape *shape = &t->getGlyphShapeTable()[i];
						graphics->currentShapeIsCharacter();
						graphics->processShape(shape,t->getFontID(),NULL);
						bitmaps.push_back(graphics->getRaster());
					}
					localFontManager->loadGlyphFaces(fontId,&bitmaps);
					list<UI8*>::iterator it;
					for (it = bitmaps.begin(); it != bitmaps.end(); it++) {
						delete []*it;
						*it = NULL;
					}
					bitmaps.clear();
					graphics->currentShapeIsNotCharacter();
				}
				if(debug) wcout << L"DefineFont2 "<< t->getFontID() << L"\n";
			}	
			break;
		}
		case stagExportAssets: 	{
			decoded = true;
			ExportAssetsTag * t = (ExportAssetsTag*)tag;
			processExportAsset(t);
			break;
		}
		case stagImportAssets: 			break;
		case stagEnableDebugger: 	 	break;
		case stagDoInitAction: 	{
			decoded = true;
			DoInitActionTag *t = (DoInitActionTag*) tag;
			if (!t->getProcessed()) doInitActiontag(t);
		 	break;
		}
		case stagDefineVideoStream: {
			decoded = true;
			DefineVideoStreamTag* t = (DefineVideoStreamTag*)tag;
			wstring videoName = encodeCharacterName(t->getCharacterID());
			ASVideo *video = (ASVideo*)objectStore->getGlobalVariable(videoName);
			if (video == NULL) {
				video = new ASVideo(script,graphics);
				video->init(t);
				video->setCharacterId(t->getCharacterID());
				objectStore->setGlobalVariable(videoName,video);
				if (debug) wcout << L"+VideoStream " << videoName << L"\n";
			}
			break;
		}
		case stagVideoFrame: {
			decoded = true;
			VideoFrameTag* t = (VideoFrameTag*)tag;
			wstring videoName = encodeCharacterName(t->getStreamID());
			ASVideo *video = (ASVideo*)objectStore->getGlobalVariable(videoName);
			video->decodeFrame(t->getFrameNum(),t->getVideoData(),t->getSize());
			break;
		}
		case stagDefineFontInfo2: {
			decoded = true;
			DefineFontInfo2Tag *t = (DefineFontInfo2Tag*) tag;
			if (!texture->characterDefined(t->getFontID())) {
				if (localFontManager->getChararacterTexture(t->getFontID(),0) == NULL) {
					localFontManager->loadFontFaces((char*)t->getFontName(),t->getFontID(),t->getFontFlagsBold(),
						t->getFontFlagsItalic(),t->getCodeTableLen(),t->getCodeTable());
				}
			}
	 		break;
		}
		case stagEnableDebugger2: 	 	break;
		case stagScriptLimits: 			break;
		case stagSetTabIndex: 				break;
		default: {
			printf("TAG code non recognized %d\n",code);
			exit(0);
		}
	}		
	if (!decoded) wcout << "Tag: "<< code << " non processed\n";
}

/**
 * Action dispatcher
 **/
void SwfEngine::actionDispatch(UI8 **data) {
	// Construtor don't have a return statement. We need to check if we have reached the end of the function code.
	int actionCode = **data;
	UI8 *addr = *data;
	if (script->checkImplicitReturn(data)) {
		if (debug) printf("[%x]Action: Implicit return code%x\n",addr,actionCode);
		script->_doReturn();
		script->popImplicitReturn();
		*data = script->popImplicitReturnAddr();
		if (actionCode == ActionPop) script->pop();
		actionCode = **data;
		if (actionCode == 0) return;
	}
	(*data)++;
	addr = *data;
	int dataLen = 0;
	if ((actionCode>>7) != 0) {
		dataLen = (UI16) **data | ((UI16) *(*data+1) << 8);
		(*data)+=2;
	}
	if (!executeMode) return;
//BUG ?	if (actionSkip) {actionSkip--; return ; }
	if (debug) printf("[%x]Action: %x ",addr,actionCode);

	switch (actionCode) {
		case ActionGotoFrame: 		script->doActionGotoFrame(data); break;
		case ActionGetURL: 			script->doActionGetURL(data);break;
		case ActionNextFrame: 		script->doActionNextFrame(); break;
		case ActionPrevFrame: 		script->doActionPrevFrame(); break;
		case ActionPlay: 			script->doActionPlay(); break;
		case ActionStop: 			script->doActionStop();  break;
		case ActionToggleQuality: 	script->doActionToggleQuality(); break;
		case ActionStopSounds: 		script->doActionStopSounds(); break;
		case ActionWaitForFrame: 	script->doActionWaitForFrame(data); break;
		case ActionSetTarget: 		script->doActionSetTarget(data); break;
		case ActionGoToLabel: 		script->doActionGoToLabel(data);break;
		case ActionPush: 			script->doActionPush(data,dataLen);break;
		case ActionPop: 			script->doActionPop(data);break;
		case ActionAdd: 			script->doActionAdd(); break;
		case ActionSubstract: 		script->doActionSubstract(); break;
		case ActionMultiply: 		script->doActionMultiply(); break;
		case ActionDivide: 			script->doActionDivide(); break;
		case ActionEqual: 			script->doActionEqual() ;break;
		case ActionLess: 			script->doActionLess() ; break;
		case ActionAnd: 			script->doActionAnd(); break;
		case ActionOr: 				script->doActionOr(); break;break;
		case ActionNot:  			script->doActionNot(); break;
		case ActionStringEquals:	script->doActionStringEquals(); break;
		case ActionStringLength:	script->doActionStringLength(); break;
		case ActionStringAdd: 		script->doActionStringAdd(); break;
		case ActionStringExtract:	script->doActionStringExtract(); break;
		case ActionStringLess: 		script->doActionStringLess(); break;
		case ActionMBStringLength:	script->doActionMBStringLength(); break;
		case ActionMBStringExtract: script->doActionMBStringExtract(); break;
		case ActionToInteger: 		script->doActionToInteger(); break;
		case ActionCharToAscii: 	script->doActionCharToAscii(); break;
		case ActionAsciiToChar: 	script->doActionAsciiToChar(); break;
		case ActionMBCharToAscii: 	script->doActionMBCharToAscii(); break;
		case ActionMBAsciiToChar: 	script->doActionMBAsciiToChar(); break;
		case ActionJump: 			script->doActionJump(data); break;
		case ActionIf: 				script->doActionIf(data); break;
		case ActionCall: 			script->doActionCall(); break;
		case ActionGetVariable: 	script->doActionGetVariable(); break;
		case ActionSetVariable: 	script->doActionSetVariable();break;
		case ActionGetURL2: 		script->doActionGetURL2(data); break;
		case ActionGotoFrame2: 		script->doActionGotoFrame2(data); break;
		case ActionSetTarget2: 		script->doActionSetTarget2(); break;
		case ActionGetProperty: 	script->doActionGetProperty(); break;
		case ActionSetProperty: 	script->doActionSetProperty(); break;
		case ActionCloneSprite: 	script->doActionCloneSprite(); break;
		case ActionRemoveSprite: 	script->doActionRemoveSprite() ; break;
		case ActionStartDrag: 		script->doActionEndDrag() ; break;
		case ActionEndDrag: 		script->doActionEndDrag() ; break;
		case ActionWaitForFrame2: 	script->doActionWaitForFrame2(data); break;
		case ActionTrace: 			script->doActionTrace();break;
		case ActionGetTime: 		script->doActionGetTime(); break;
		case ActionRandomNumber: 	script->doActionRandomNumber(); break;
		case ActionCallFunction: 	script->doActionCallFunction(data); break;
		case ActionCallMethod: 		script->doActionCallMethod(data);break;
		case ActionConstantPool: 	script->doActionConstantPool(data); break;
		case ActionDefineFunction: 	script->doActionDefineFunction(data) ;  break;
		case ActionDefineLocal: 	script->doActionDefineLocal() ;break;
		case ActionDefineLocal2:	script->doActionDefineLocal2() ;break;
		case ActionDelete: 			script->doActionDelete(); break;
		case ActionDelete2: 		script->doActionDelete2();break;
		case ActionEnumerate: 		script->doActionEnumerate() ; break;
		case ActionEquals2:		 	script->doActionEquals2(); break;
		case ActionGetMember: 		script->doActionGetMember(); break;
		case ActionInitArray: 		script->doActionInitArray(); break;
		case ActionInitObject: 		script->doActionInitObject(); break;
		case ActionNewMethod: 		script->doActionNewMethod(data); break;
		case ActionNewObject: 		script->doActionNewObject(data); break;
		case ActionSetMember: 		script->doActionSetMember(); break;
		case ActionTargetPath:		script->doActionTargetPath(); break;
		case ActionWith: 			script->doActionWith(data) ; break;
		case ActionToNumber: 		script->doActionToNumber(); break;
		case ActionToString: 		script->doActionToString() ; break;
		case ActionAdd2: 			script->doActionAdd2(); break;
		case ActionLess2:  			script->doActionLess2(); break;
		case ActionModulo:			script->doActionModulo();break;
		case ActionBitAnd: 			script->doActionBitAnd();break;
		case ActionBitLShift: 		script->doActionBitLShift(); break;
		case ActionBitOr: 			script->doActionBitOr(); break;
		case ActionBitRShift: 		script->doActionBitRShift(); break;
		case ActionBitURShift: 		script->doActionBitURShift(); break;
		case ActionBitXor: 			script->doActionBitXor(); break;
		case ActionDecrement:  		script->doActionDecrement(); break;
		case ActionIncrement:  		script->doActionIncrement(); break;
		case ActionPushDuplicate: 	script->doActionPushDuplicate(); break;
		case ActionReturn: 			script->doActionReturn(data); break;
		case ActionStackSwap: 		script->doActionStackSwap(); break;
		case ActionStoreRegister: 	script->doActionStoreRegister(data) ; break;
		case DoInitAction: 			script->doDoInitAction() ; break;
		case ActionInstanceOf: 		script->doActionInstanceOf(); break;
		case ActionEnumerate2: 		script->doActionEnumerate2() ; break;
		case ActionStrictEquals: 	script->doActionStrictEquals(); break;
		case ActionGreater: 		script->doActionGreater(); break;
		case ActionStringGreater: 	script->doActionStringGreater() ; break;
		case ActionDefineFunction2:	script->doActionDefineFunction2(data); break;
		case ActionExtends:	 		script->doActionExtends(); break;
		case ActionCastOp: 			script->doActionCastOp(); break;
		case ActionImplementsOp: 	script->doActionCastOp() ; break;
		case ActionTry: 			script->doActionTry() ; break;
		case ActionThrow: 			script->doActionThrow() ; break;
		default: {
			wcout << L"ACTION code non recognized Code:" << actionCode << L"\n";
		}
	}
	if (debug) wcout << L"\n";
}

/**
 * Execute Show Frame 
 * Process pending doAction tags
 **/
void SwfEngine::doShowFrame() {
//puts("doShowFrame");
	// Process first sprite level and call innerDisplay to recurse if necessary.
	IntScriptObjectMap *movies = script->getObjectStore()->getObjectByDepth();
	for (IntScriptObjectMap::iterator it=movies->begin(); it != movies->end(); it++) {
		script->setCurrentContext((ScriptObject*)objectStore->getGlobalVariable(L"_global"));
		ASObject *aos = (ASObject *)(*it).second;
		if (aos == NULL) continue;
		if (isSpriteObject(aos)) {
			innerDisplay((ASSprite*)aos);
		}
	}

	if (!executeMode) return;
	onEnterMainFrame();

	if (doActionList.size() >0) {
		for (list<DoActiontag*>::iterator it = doActionList.begin(); !actionListChanged && it != doActionList.end() ; it++) {
			DoActiontag*tag = *it;
			if (tag->getCurSpriteId() == 0) {
				ScriptObject *global = (ScriptObject*)objectStore->getGlobalVariable(L"_global");
				script->setCurrentContext(global);
				objectStore->targetPathDown(global);
				UI8 *code = tag->getCode();
				noActionScript = true;
				while (*code != 0) {
					actionDispatch(&code);
				}
				noActionScript = false;
				objectStore->targetPathUp();
			} else {
				// The script applies for all clones sprites.
				list<ScriptObject*> listClones = objectStore->getClonedObjects(tag->getCurSpriteId());
				for (list<ScriptObject*>::const_iterator it = listClones.begin(); it != listClones.end(); ++it) {
					ASObject *pobj = (ASObject*)*it;
					if (pobj->getIsStopped()) continue;
					objectStore->targetPathDown(pobj);
					script->setCurrentContext(pobj);
					UI8 *code = tag->getCode();
					noActionScript = true;
					while (*code != 0) {
						actionDispatch(&code);
					}
					noActionScript = false;
					objectStore->targetPathUp();
				}
			}	
		}
		doActionList.clear();
	}
	actionListChanged = false;
	eventManager->resetLastKeyPressed();
}

/**
 * Go into Sprite hierachy. 
 * This func may call itself for sprites than contains sprites.
 **/
void SwfEngine::innerDisplay(ASSprite *sprite) {
	script->actionContextPush(sprite,sprite->getMovieName());
	continueDecode = !sprite->getIsStopped(); 
	if (continueDecode) {
		if (sprite->getCurrentTagType() == stagShowFrame) sprite->goNextTag(true);
		if (sprite->isAtEnd()) {
			if (sprite->getFrameCount() > 1) {
				sprite->actionGotoFrame(0);
				sprite->clearCharacters();
			}
			else continueDecode = false;
		}
		while (continueDecode) {
			if (sprite->getCurrentTagType() == stagEnd) {
				continueDecode = false;
				sprite->actionGotoFrame(0);
				continue;
			}
			if ( sprite->getCurrentTag() == NULL || sprite->getCurrentTagType() == stagShowFrame) {
				continueDecode = false;
			} else {
				tagDispatch(*(sprite->getCurrentTag()));
				sprite->goNextTag();
			}
		}
	
		for (IntASObjectMap::iterator it = sprite->getCharacters()->begin(); it !=sprite->getCharacters()->end(); it++) {
			ASObject *obj = (*it).second;
			if (isSpriteObject(obj)) {
				innerDisplay((ASSprite*)obj);
			}
		}
	} else {
		//if (debug) wcout << sprite->getMovieName() << L" stopped\n";
	}
	script->actionContextPop();
}

/**
 * Execute place object
 **/
void SwfEngine::doPlaceObject2(PlaceObject2Tag *tag) {
	int depth = tag->getDepth();
	int spriteId = tag->getCurSpriteId();
	bool	hadTransformMatrix,hadCxFormWithAlpha,hadRemove=false;
	MATRIX  oldtransformMatrix;
	CXFORMWITHALPHA oldCxFormWithAlpha;
	int 	clipDepth;
	//printf("PLACE OBJECT2 : %d %d %d",depth,spriteId,tag->getCharacterId());
	
	ASObject *dobj = NULL;
	if (tag->getPlaceFlagHasCharacter()) dobj = (ASObject *) objectStore->getGlobalVariable(encodeCharacterName(tag->getCharacterId()));

	if (spriteId != 0) {
		ASSprite *parent = (ASSprite*)objectStore->getGlobalVariable(encodeCharacterName(spriteId));
		if (tag->getPlaceFlagMove() && tag->getPlaceFlagHasCharacter()) {
			hadTransformMatrix = dobj->getHasTransformMatrix();
			hadCxFormWithAlpha = dobj->getHasCxFormWithAlpha();
			if (hadTransformMatrix) oldtransformMatrix = *dobj->getTransformMatrix();
			if (hadCxFormWithAlpha) oldCxFormWithAlpha = *dobj->getCxFormWithAlpha();
			clipDepth = dobj->getClipDepth();
			hadRemove = true;
			ASInterface *o = parent->getCharacter(depth);
			if (o != NULL) o->setReferenced(false);
			parent->removeCharacter(depth);
		}

		if (tag->getPlaceFlagHasName()) {
			wstring s = tag->getName()->getString();
			objectStore->setGlobalVariable(s,dobj);		// Object now referencable by it's name
			list<ScriptObject*>clones = objectStore->getClonedObjects(dobj->getCharacterId());
			for (list<ScriptObject*>::const_iterator i=clones.begin(); i!= clones.end(); ++i) {
				ASObject* a = (ASObject*) *i;
				ASObject* parent = (ASObject*)a->getParent();
				if (parent != NULL) { 
					objectStore->setSpriteVariable(parent,s,a);		// Object now referencable by it's name
				}
			}
			dobj->setMovieName(s);
		} 

		if (tag->getPlaceFlagHasCharacter()) {
			dobj->setVisible(true);
			if (parent->getCharacter(depth) == NULL ) {
				if (dobj->getIsReferenced()) {
					dobj = (ASObject *) objectStore->getGlobalVariable(encodeCharacterName(tag->getCharacterId()));
					//printf("CLONE1: %d %d %d %x\n",depth,tag->getCharacterId(),spriteId,dobj);
					dobj = (ASObject*)dobj->getObjectDeepCopy(false);
				}
				//printf("Add %d %d %x to %x ",depth,spriteId,dobj,sprite);wcout << encodeCharacterName(spriteId) << L"\n";
			}
			wstring s = encodeCharacterName(tag->getCharacterId());
			objectStore->setSpriteVariable(parent,s,dobj);	
			parent->addCharacter(depth,dobj);
			dobj->setReferenced(true);		// The sprite/movie clip is referenced, so we need to copy it if it's referenced an other time

			// Restore previous placement setting setting
			if (hadRemove) {
				dobj->setHasTransformMatrix(hadTransformMatrix);
				dobj->setHasCxFormWithAlpha(hadCxFormWithAlpha);
				if (hadTransformMatrix) dobj->setTransformMatrix(&oldtransformMatrix);
				if (hadCxFormWithAlpha) dobj->setCxFormWithAlpha(&oldCxFormWithAlpha);
				dobj->setClipDepth(clipDepth);
			}
		}

		dobj = parent->getCharacter(depth);
		dobj->setParent(parent);
		if (tag->getPlaceFlagHasMatrix() ){
			dobj->setTransformMatrix(tag->getMatrix());
			dobj->setHasTransformMatrix(true);
		}

		if (tag->getPlaceFlagHasColorTransform()){
			dobj->setCxFormWithAlpha(tag->getColorTransform());
			dobj->setHasCxFormWithAlpha(true);
		}

		if (tag->getPlaceFlagHasClipDepth()) {
			dobj->setClipDepth(tag->getClipDepth()); 
		} else {
			dobj->setClipDepth(0);
		}

		parent->setVisible(true);
		if (tag->getPlaceFlagHasClipActions()) {
			eventManager->registerClipAction(dobj,tag->getClipActions());
		}
	} else  {
		dobj = (ASObject *) script->getObjectStore()->getObjectAtDepth(depth);
		if (tag->getPlaceFlagMove() && tag->getPlaceFlagHasCharacter()) {
			hadTransformMatrix = dobj->getHasTransformMatrix();
			hadCxFormWithAlpha = dobj->getHasCxFormWithAlpha();
			if (hadTransformMatrix) oldtransformMatrix = *dobj->getTransformMatrix();
			if (hadCxFormWithAlpha) oldCxFormWithAlpha = *dobj->getCxFormWithAlpha();
			clipDepth = dobj->getClipDepth();
			hadRemove = true;
			ASInterface *o = (ASInterface *)script->getObjectStore()->getObjectAtDepth(depth);
			if (o != NULL) o->setReferenced(false);
			script->getObjectStore()->removeObjectAtDepth(depth);
		}

		if (tag->getPlaceFlagHasCharacter()) {
			if (script->getObjectStore()->getObjectAtDepth(depth) == NULL) { 
				dobj = (ASObject *) objectStore->getGlobalVariable(encodeCharacterName(tag->getCharacterId()));
				dobj->setVisible(true);
				if (dobj->getIsReferenced()) {
					dobj = (ASObject*)dobj->getObjectDeepCopy(true);
					//printf("CLONE2: %d %d %d  %x\n",depth,tag->getCharacterId(),spriteId,dobj);
				} else {
					dobj->setReferenced(true);
				}
				script->getObjectStore()->setObjectAtDepth(depth,dobj);
			
				// Restore previous placement setting setting
				if (hadRemove) {
					dobj->setHasTransformMatrix(hadTransformMatrix);
					dobj->setHasCxFormWithAlpha(hadCxFormWithAlpha);
					if (hadTransformMatrix) dobj->setTransformMatrix(&oldtransformMatrix);
					if (hadCxFormWithAlpha) dobj->setCxFormWithAlpha(&oldCxFormWithAlpha);
					dobj->setClipDepth(clipDepth);
				}
			}
		} 
		if (dobj == NULL) {
			puts("Object deleted, but still placed!");
			return;
		}

		if (tag->getPlaceFlagHasMatrix()) {
			bool update = true;
			if (isSpriteObject(dobj)) {
				ASSprite *s = (ASSprite*)dobj;
				if (s->getModifiedByScript()) update = false;
			}
			if (update) {
				dobj->setTransformMatrix(tag->getMatrix());
				dobj->setHasTransformMatrix(true);
			}
		}

		if (tag->getPlaceFlagHasColorTransform()) {
			dobj->setCxFormWithAlpha(tag->getColorTransform());
			dobj->setHasCxFormWithAlpha(true);
		}

		if (tag->getPlaceFlagHasClipDepth()) {
			dobj->setClipDepth(tag->getClipDepth());
		}

		if (tag->getPlaceFlagHasRatio()) {
			ASMovieClip *m = (ASMovieClip*)dobj;
			m->setRatio(float(tag->getRatio())/65536.0f);
		}

		if (tag->getPlaceFlagHasName()) {
			wstring s = tag->getName()->getString();
			list<ScriptObject*>clones = objectStore->getClonedObjects(dobj->getCharacterId());
			for (list<ScriptObject*>::const_iterator i=clones.begin(); i!= clones.end(); ++i) {
				ASObject* a = (ASObject*) *i;
				ASObject* parent = (ASObject*)a->getParent();
				if (parent != NULL) { 
					objectStore->setSpriteVariable(parent,s,a);		// Object now referencable by it's name
				}
			}
			dobj->setMovieName(s);
			objectStore->setGlobalVariable(s,dobj);
		} 
		dobj->setVisible(true);

		// If we have clipActions, we register them. 
		if (tag->getPlaceFlagHasClipActions()) {
			eventManager->registerClipAction(dobj,tag->getClipActions());
		}
	}
}

/**
  * Goto Frame
 **/
void SwfEngine::goToFrame(int frame,list<BaseTag*> *tagList) {
}

/**
 * Go to Frame on the main track (_global object)
 **/
void SwfEngine::goToMainTrackFrame(int iframe) {
	script->getObjectStore()->clear();
	eventManager->clearInteractions();
	mainTag = tagList->begin();
	doActionList.clear();
	actionListChanged = true;
	executeMode = false;
	noActionScript = true;
	currentFrame = 0;
	for (int i=0; i< iframe; i++) next();
	executeMode = true;
	noActionScript = false;
}

/**
 * Go to previous frame on _global
 **/
void  SwfEngine::actionPreviousFrame() {
	int targetFrame = currentFrame > 0 ? currentFrame -1 : 0;
	goToMainTrackFrame(targetFrame);
}


/**
 * Mouse handler for clickdoOnLoad event
 **/
bool SwfEngine::mouseClickCallback(int x, int y,int mouseAction) {
	return eventManager->mouseClickCallback(x,y,mouseAction);
}

/**
 * Keyboard handler
 **/
void SwfEngine::keyboardCallback(int key, char *code) {
	eventManager->keyboardCallback(key,code);
}

/**	
 * Process exportAsset.
 * The flash file is scanned for a character ID corresponding to an export asset.
 * The tag is processed (a doInitAction tag) to register actions.
 **/
void SwfEngine::processExportAsset(ExportAssetsTag *t) {
	for (int i=0; i< t->getCount() ; i++) {
		wstring name  = t->getNames()[i]->getString();
		int characterId = t->getTags()[i];
		objectStore->addExportedAsset(name,characterId);
	}
} 

/**
 * DoInitAction Tag (Flash 6)
 **/
void SwfEngine::doInitActiontag(DoInitActionTag *tag) {
	puts("doInitActiontag");
	tag->setProcessed(true);
	// We have an extra pop() in the class init sequence.
	printf("BASE TAG CODE:%x\n",tag->getCode());
	script->pushExtraValue();
	script->pushExtraValue();
	script->pushExtraValue();
	script->pushExtraValue();
	script->pushExtraValue();
	UI8 *code = tag->getCode();
	while (*code != 0) actionDispatch(&code);
	objectStore->setGlobalActionContext();
	objectStore->setPoolActionContext(L"_global");
}

/**
 * Return if the object is of Sprite class
 **/
bool SwfEngine::isSpriteObject(ASInterface *a) {
	if (a->getClassName() != L"Sprite") return false;
	ASSprite *as = (ASSprite*) a;
	return as->getCharacters()->size() != 0;
}

/**
 * This function is called when we enter a new frame
 **/
void SwfEngine::onEnterMainFrame() {
	if (!executeMode) return;
	eventManager->doOnLoad();
	eventManager->doTimerCallbacks();
	movieEnterFrameEvent();
}

/**
 * Loop over movies to call the OnEnterFrame event
 **/
void SwfEngine::movieEnterFrameEvent() {
	// Process global events
	eventManager->doOnEnterFrame((ScriptObject*)objectStore->getGlobalVariable(L"_global"));
	IntScriptObjectMap *movies = script->getObjectStore()->getObjectByDepth();
	for (IntScriptObjectMap::iterator it=movies->begin(); it != movies->end(); ++it) {
		ASObject *pobj = (ASObject*)(*it).second;
		if (pobj == NULL) continue;
		script->setCurrentContext(pobj);
		if (isSpriteObject(pobj)) {
			spriteEnterFrameEvent((ASSprite*)pobj);
		} 
		if (pobj->getThisObjectInstanceName() == L"Button") {
			// TODO 
			// Process button onFrameEvent
			//movieEnterFrameEvent((ASButton *)amc,currentDepth);
		} else {
			objectStore->targetPathDown(pobj);
			eventManager->doOnEnterFrame(pobj);
			objectStore->targetPathUp();
		}
	}
}

/**
  * Recursive calls to character rendering in sprites
  **/
void SwfEngine::spriteEnterFrameEvent(ASSprite *sprite) {
	if (sprite->getIsStopped()) return;
	for (IntASObjectMap::iterator it=sprite->getCharacters()->begin(); it != sprite->getCharacters()->end(); it++) {
		ASObject *pobj = (*it).second;
		if (isSpriteObject(pobj)) {
			spriteEnterFrameEvent((ASSprite*)pobj);
		} 
		objectStore->targetPathDown(pobj);
		eventManager->doOnEnterFrame(pobj);
		objectStore->targetPathUp();
		if (pobj->getThisObjectInstanceName() == L"Button") {
			// TODO 
			// Process button onFrameEvent
		}
	}
	objectStore->targetPathUp();
}

/**
 * Remove a movieClip in global or in a sprite
 **/
void SwfEngine::removeObject(RemoveObject2Tag *t){
	//printf("REMOVE....................%d\n",t->getDepth());

	// Unregister events.
	ASObject *obj = (ASObject *)script->getObjectStore()->getObjectAtDepth(t->getDepth());
	eventManager->unRegisterClipAction(obj);

	if (t->getCurSpriteId() ==0 ) {
		IntScriptObjectMap *movies = script->getObjectStore()->getObjectByDepth(); 
		int depth  = movies->find(t->getDepth())->first;
		ASInterface *a  = (ASInterface *)script->getObjectStore()->getObjectAtDepth(depth);
		a->setClipDepth(0);
		a->setReferenced(false);
		movies->erase(depth);
		script->getObjectStore()->removeObjectAtDepth(t->getDepth());

		// Remove access by tagId name or nickname (MovieName)
		//??script->getObjectStore()->removeGlobalVariable(obj->getMovieName());
		//??script->getObjectStore()->removeGlobalVariable(encodeCharacterName(obj->getCharacterId()));
	} else {
		// remove associated actions
		list<DoActiontag*>::iterator itRemove = NULL;
		for (list<DoActiontag*>::iterator it = doActionList.begin(); it != doActionList.end() ; ++it) {
			DoActiontag *t = *it;
			if (t->getCurSpriteId() == t->getCurSpriteId()) itRemove = it; 
		}
		if (itRemove != NULL) doActionList.erase(itRemove);

		// Remove access by tagId name or nickname (MovieName)
		ASSprite* parent = (ASSprite *)script->getObjectStore()->getGlobalVariable(encodeCharacterName(t->getCurSpriteId()));
		obj = parent->getCharacter(t->getDepth());
		//?? parent->removeCharacter(t->getDepth());
		//??script->getObjectStore()->removeSpriteVariable(parent,obj->getMovieName());
		//??script->getObjectStore()->removeSpriteVariable(parent,encodeCharacterName(obj->getCharacterId()));
	}
	//?? script->getObjectStore()->destroyObject(obj);	// Calls the destructor
}
